from .history import History, HistoryPlotter, HistoryWriter
from .logging import Logger
from .timing import Timing
from .base import FunctionCallback
from .eval import MulticlassEval
from .checkpoint import Checkpoint
from .stop import StopOnNaN


from yann.callbacks.base import Callback
from yann import set_param
from math import cos, pi


def cosine_anneal(min_lr, max_lr, cur_step, num_steps):
  return min_lr + (max_lr - min_lr) * (1 + cos(cur_step / num_steps * pi)) / 2


class SGDR(Callback):
  def __init__(
      self,
      optimizer=None,
      max_lr=None,
      min_lr=0,
      cycle_len=10,
      cycle_mult=1,
      lr_mult=1,
      verbose=True,
  ):
    self.optimizer = optimizer
    self.max_lr = max_lr
    self.min_lr = min_lr
    self.base_cycle_len = cycle_len
    self.cycle_mult = cycle_mult
    self.lr_mult = lr_mult

    self.cur_cycle_len = self.base_cycle_len
    self.cur_max_lr = self.max_lr
    self.cur_min_lr = self.min_lr
    self.cur_lr = self.max_lr
    self.cur_step = 0

    self.completed_cycles = 0

    self.verbose = verbose

    if self.cur_max_lr and self.optimizer:
      self.update_lr(self.cur_lr)

  def restart(self):
    self.cur_cycle_len *= self.cycle_mult
    self.cur_max_lr *= self.lr_mult
    self.cur_min_lr *= self.lr_mult

    self.cur_lr = self.cur_max_lr
    self.cur_step = 0
    self.update_lr(self.cur_lr)

    if self.verbose:
      print('Restarting SGDR')

  def update_lr(self, lr):
    set_param(self.optimizer, 'lr', lr)

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    if self.cur_step >= self.cur_cycle_len:
      self.restart()
    else:
      new_lr = cosine_anneal(self.cur_min_lr, self.cur_max_lr, self.cur_step,
                             self.cur_cycle_len)
      self.update_lr(new_lr)
      self.cur_step += 1



import subprocess

from yann.callbacks.base import Callback

def gcp_sync(src, dst, exclude=None):
  if exclude:
    return subprocess.call([
      'gsutil',
      '-m',
      'rsync',
      '-r',
      '-x',
      exclude,
      src,
      dst
    ])
  else:
    return subprocess.call([
      'gsutil',
      '-m',
      'rsync',
      '-r',
      src,
      dst
    ])



class SyncCallback(Callback):
  def upload(self):
    pass

  def download(self):
    pass


class GCPSync(SyncCallback):
  pass


from torch.utils.data.sampler import Sampler
import random
from collections import defaultdict


class BalancedTargetSampler(Sampler):
  def __init__(self, dataset):
    super().__init__(dataset)
    self.dataset = dataset

    self.target_to_indices = defaultdict(list)
    for n, targets in enumerate(self.dataset.targets):
      for t in targets:
        self.target_to_indices[t].append(n)
    self.targets = list(self.target_to_indices)

  def __iter__(self):
    for n in range(len(self)):
      t = random.choice(self.targets)
      yield random.choice(self.target_to_indices[t])

  def __len__(self):
    return len(self.dataset)


import torch.nn.functional as F
import torch
from torch.nn.modules.loss import _Loss


def soft_target_cross_entropy(input, target, reduce=True,
                              reduction='elementwise_mean'):
  """"like cross_entropy but using soft tarets"""
  vals = torch.sum(-target * F.log_softmax(input, dim=1))
  if not reduce:
    return vals

  if reduction == 'mean':
    return torch.sum(vals)
  elif reduction == 'elementwise_mean':
    return torch.mean(vals)
  elif 'none':
    return vals
  raise ValueError('Unsupported reduction mode: ' + str(reduction))



class SoftTargetCrossEntopyLoss(_Loss):
  def __init__(self, reduce=None, reduction='elementwise_mean'):
    super().__init__(reduce=reduce, reduction=reduction)
    self.reduce = reduce
    self.reduction = reduction

  def forward(self, input, target):
    return soft_target_cross_entropy(input, target, reduce=self.reduce,
                                     reduction=self.reduction)



from typing import Optional

import torch
from torch.utils.data import DataLoader
import torchvision
import logging
import datetime
import pathlib
import numpy as np
import datetime


from .base import BaseTrainer
from yann.data import TransformDataset
from yann import resolve, evaluate
from yann import callbacks as yann_callbacks


def timestr():
  return f"{datetime.datetime.utcnow().strftime('%y-%m-%dT%H%M%S')}"


class Trainer(BaseTrainer):
  model: torch.nn.Module
  lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler]
  loader: DataLoader

  def __init__(
      self,
      model,
      dataset=None,
      optimizer=None,
      loss=None,
      loader=None,
      sampler=None,
      num_workers=8,
      transform=None,
      lr_scheduler=None,
      callbacks=None,
      device=None,
      trainable_parameters=None,
      batch_size=16,
      val_dataset=None,
      val_loader=None,
      val_transform=None,
      parallel=False,
      name=None,
      description=None,
      root='./'
  ):
    super().__init__()

    self.model = model
    if parallel:
      self.model = torch.nn.DataParallel(self.model)

    self.loss = resolve(
      loss,
      (torch.nn, torch.nn.functional),
      required=True,
      validate=lambda x: callable(x)
    )
    self.optimizer = resolve(
      optimizer,
      (torch.optim,),
      required=True,
      validate=lambda x: hasattr(x, 'step'),
      params=trainable_parameters or self.model.parameters()
    )
    self.dataset = resolve(
      dataset,
      (torchvision.datasets,),
      required=True,
    )

    if transform:
      self.dataset = TransformDataset(self.dataset, transform)

    self.batch_size = batch_size

    self.loader = loader or DataLoader(
      self.dataset,
      batch_size=self.batch_size,
      pin_memory=True,
      shuffle=False if sampler else True,
      sampler=sampler,
      #       collate_fn=collate,
      num_workers=num_workers
    )

    self.val_dataset = val_dataset

    val_transform = val_transform or transform
    if val_transform:
      self.val_dataset = TransformDataset(self.val_dataset, val_transform)

    self.val_loader = val_loader or (val_dataset and DataLoader(
      self.val_dataset,
      batch_size=batch_size,
      shuffle=False,
      pin_memory=True,
      num_workers=num_workers
    ))

    self.lr_scheduler = resolve(
      lr_scheduler,
      (torch.optim.lr_scheduler,),
      optimizer=self.optimizer
    )

    self.num_samples = 0
    self.num_steps = 0
    self.num_epochs = 0

    self.name = name or (
      f"{timestr()}-{self.model.__class__.__name__}"
    )
    self.description = description

    self.root = pathlib.Path(root) / self.name
    self.root.mkdir(parents=True, exist_ok=True)

    self.history = None
    has_history = False
    if callbacks:
      for c in callbacks:
        if isinstance(c, yann_callbacks.History):
          self.history = c
          has_history = True
          break
    self.history = self.history or yann_callbacks.History()

    self.callbacks = callbacks or [
      yann_callbacks.Logger()
    ]

    if not has_history:
      self.callbacks.append(self.history)

    # make sure history callback is called first
    self.callbacks = sorted(
      self.callbacks,
      key=lambda x: 0 if x is self.history else 1)

    self.function_callback = None

    if device:
      self.device = torch.device(device) if isinstance(device, str) else device
      self.to(self.device)
    else:
      self.device = None

  def to(self, device=None):
    self.device = device

    if self.model:
      self.model.to(self.device)

  def on(self, event, callback):
    if not self.function_callback:
      self.function_callback = yann_callbacks.FunctionCallback()
      self.callbacks.append(self.function_callback)

    self.function_callback.on(event, callback)

  def step(self, inputs, target):
    self.model.train()
    self.optimizer.zero_grad()

    outputs = self.model(inputs)
    loss = self.loss(outputs, target)
    loss.backward()
    self.optimizer.step()

    self.num_steps += 1
    self.num_samples += len(inputs)

    return outputs, loss

  def batches(self):
    for inputs, targets in self.loader:
      yield inputs.to(self.device), targets.to(self.device)

  def validate(self, loader=None, device=None):
    loader = loader or self.val_loader
    device = device or self.device

    self.model.eval()

    self.on_validation_start()

    ts, os = [], []
    with torch.no_grad():
      for inputs, targets, outputs in evaluate(self.model, loader, device):
        self.on_validation_batch(inputs=inputs, targets=targets, outputs=outputs)
        ts.append(targets)
        os.append(outputs)

      ts = torch.cat(ts)
      os = torch.cat(os)
      loss = self.loss(os, ts)
    self.on_validation_end(targets=ts, outputs=os, loss=loss)
    return loss

  def run(self, epochs=1):
    self._stop = False
    try:
      self.on_train_start()

      for _ in range(epochs):
        self.on_epoch_start(epoch=self.num_epochs)
        for batch_idx, (inputs, targets) in enumerate(self.batches()):
          self.on_batch_start(
            batch=batch_idx,
            inputs=inputs,
            targets=targets
          )
          try:
            outputs, loss = self.step(inputs, targets)
          except KeyboardInterrupt as e:
            self.stop()
            break
          except Exception as e:
            self.on_batch_error(batch=batch_idx, error=e)
            if self._stop: break
            raise e

          self.on_batch_end(batch=batch_idx, inputs=inputs, targets=targets,
                            outputs=outputs, loss=loss)
          if self._stop: break
        if self._stop: break

        if self.val_loader:
          val_loss = self.validate()
          if self.lr_scheduler:
            self.lr_scheduler.step(val_loss)
        elif self.lr_scheduler:
          self.lr_scheduler.step()

        self.on_epoch_end(epoch=self.num_epochs)
        self.num_epochs += 1

      self.on_train_end()
    except Exception as e:
      self.on_error(e)
      raise e

  def __call__(self, *args, **kwargs):
    self.run(*args, **kwargs)

  def stop(self):
    self._stop = True

  def get_checkpoint_name(self):
    return (
      f"{self.name}-steps_{self.num_steps}.th"
    )

  @property
  def checkpoints_root(self):
    root = self.root / 'checkpoints'
    root.mkdir(parents=True, exist_ok=True)
    return root

  def checkpoint(self, root='./'):
    state = self.state_dict()
    torch.save(
      state,
      str(self.checkpoints_root / f"{timestr()}-epoch-{self.num_epochs:03d}.th")
    )

  def load_checkpoint(self, path, metadata=True):
    data = torch.load(path)
    self.load_state_dict(data, metadata=metadata)

  def state_dict(self):
    data = {
      'metadata': {
        'num_steps': self.num_steps,
        'num_samples': self.num_samples,
        'num_epochs': self.num_epochs,
        'batch_size': self.batch_size,
        'name': self.name,
        'time_created': datetime.datetime.utcnow()
      }
    }

    for k, v in self.__dict__.items():
      if hasattr(v, 'state_dict'):
        data[k] = {
          'state_dict': v.state_dict(),
          'class': v.__class__.__name__
        }

    return data

  def load_state_dict(self, data, metadata=True):
    """
    TODO: add a way to specify which parts should be loaded (ex: model only)
    """
    skipped = set()
    for k, v in data.items():
      if 'state_dict' in v and hasattr(self, k):
        getattr(self, k).load_state_dict(v['state_dict'])
      else:
        skipped.add(k)

    if metadata and 'metadata' in data:
      skipped.discard('metadata')
      for k, v in data['metadata'].items():
        setattr(self, k, v)

    if skipped:
      logging.warning(f'skipped {skipped} when loading checkpoint')

  def __str__(self):
    return f"""
name: {self.name}
batch_size: {self.batch_size}
device: {self.device}

MODEL
=====

{self.model}


DATASET
=======

{self.loader.dataset}


LOADER
======

{self.loader}

LOSS
====

{self.loss}


OPTIMIZER
=========

{self.optimizer}


PROGRESS
========
epochs: {self.num_epochs}
steps: {self.num_steps}
samples: {self.num_samples}"""



from itertools import zip_longest
from torch.utils.data import Dataset
from .classes import Classes


class TransformDataset(Dataset):
  def __init__(self, dataset, transforms):
    super().__init__()
    self.dataset = dataset
    self.transforms = transforms if isinstance(transforms, tuple) else (
      transforms,)

  def __len__(self):
    return len(self.dataset)

  def __getitem__(self, idx):
    return tuple(
      t(x) if t else x
      for (x, t) in zip_longest(self.dataset[idx], self.transforms)
    )

  def __repr__(self):
    return (
      f'{self.__class__.__name__}('
      f'\nDataset: {repr(self.dataset)}'
      f'\nTransforms: {repr(self.transforms)}'
      f'\n)')


class Noisy(Dataset):
  def __init__(self, dataset, p=0.1):
    self.clean_dataset = dataset
    self.p = p

  def __len__(self):
    return len(self.clean_dataset)

  def __getitem__(self, idx):
    x, y = self.clean_dataset[idx]



class Imbalanced(Dataset):
  def __init__(self, dataset):
    self.dataset = dataset



class InputsTargetsDataset(Dataset):
  def __init__(self, inputs, targets, transform=None):
    self.inputs = inputs
    self.targets = targets
    self.transform = transform
    self.classes = Classes(sorted(set(x for t in self.targets for x in t)))

  def __len__(self):
    return len(self.inputs)

  def __getitem__(self, idx):
    x, y = self.inputs[idx], self.targets[idx]
    if self.transform:
      x = self.transform(x)
    return x, y


from .classification import Trainer


def step(model, x, y, optimizer, criterion):
  model.train()
  optimizer.zero_grad()

  pred = model(x)
  loss = criterion(pred, y)

  loss.backward()
  optimizer.step()

  return x, y, pred, loss


def train(model, batches, optimizer, criterion, device=None):
  for x, y in batches:
    if device:
      x, y = x.to(device), y.to(device)
    yield step(model, x, y, optimizer, criterion)


from io import BytesIO
from PIL import Image
from itertools import zip_longest
import base64


def base64str(img):
  s = BytesIO()
  img.save(s, format='JPEG')
  return (
      "data:image/jpeg;base64,"
      + base64.b64encode(s.getvalue()).decode("utf-8")
  )


def show_images(paths, labels=None, w=400, h=400):
  from IPython.core.display import display, HTML

  tags = []
  for x, l in zip_longest(paths, labels or []):
    img = Image.open(x) if isinstance(x, str) else x
    img.thumbnail((w, h))

    tags.append(
      f'''
      <img 
        style="max-width: {w}px; max-height: {h}px; margin: 3px;"
         src={base64str(img)} />
      '''
    )
  return display(HTML(f'<div>{"".join(tags)}</div>'))


import numpy as np
import collections
import torch

class Classes:
  valid_encodings = {
    'index',
    'one_hot',
    'normalized_one_hot'
  }

  def __init__(
      self,
      classes,
      meta=None,
      default_encoding='index'
  ):
    self.classes = list(classes)
    self.indices = {c: i for i, c in enumerate(self.classes)}
    self.meta = meta

    self.dtype = 'float32'

    assert default_encoding in self.valid_encodings, \
      f'default_encoding must be one of {self.valid_encodings}, got {default_encoding}'
    self.default_encoding = default_encoding

  def __getitem__(self, idx):
    return self.classes[idx]

  def __iter__(self):
    return iter(self.classes)

  def __contains__(self, cls):
    return cls in self.indices

  def __len__(self):
    return len(self.classes)

  def __eq__(self, other):
    return self.indices == other.indices

  def encode(self, seq, encoding=None):
    return getattr(self, (encoding or self.default_encoding) + '_encode')(seq)

  def decode(self, encoded, encoding=None):
    return getattr(self, (encoding or self.default_encoding) + '_decode')(encoded)

  def index_encode(self, classes):
    if isinstance(classes, (str, int)):
      return self.indices[classes]
    return [self.indices[c] for c in classes]

  def index_decode(self, indices):
    if isinstance(indices, int):
      return self[indices]
    return [self[idx] for idx in indices]

  def one_hot_encode(self, classes):
    if isinstance(classes, (str, int)):
      classes = [classes]
    y = np.zeros(len(self), dtype=self.dtype)
    y[[self.indices[c] for c in classes]] = 1
    return y

  def normalized_one_hot_encode(self, classes):
    if isinstance(classes, (str, int)):
      classes = [classes]
    y = np.zeros(len(self), dtype=self.dtype)
    y[[self.indices[c] for c in classes]] = 1
    y = np.true_divide(y, y.sum())
    return y

  def ranked_decode(self, scores):
    indices = np.argsort(scores)
    return [(self.classes[i], scores[i]) for i in indices[::-1]]


def smooth(y, eps=.1):
  return y * (1 - eps) + eps * (1.0 / len(y))


import torch
from yann import to_numpy


def get_preds(scores):
  score, preds = torch.max(scores, dim=1)
  return preds


def top_k(scores, k=5, largest=True):
  return torch.topk(scores, k=k, dim=1, largest=largest)


def accuracy(targets, preds):
  if targets.shape != preds.shape:
    preds = get_preds(preds)
  return (targets == preds).sum().float() / len(preds)


def evaluate_multiclass(
    targets,
    outputs,
    preds=None,
    classes=None
):
  preds = preds or get_preds(outputs)
  targets, outputs, preds = (
    to_numpy(targets),
    to_numpy(outputs),
    to_numpy(preds)
  )






def evaluate_multilabel(
    targets,
    outputs,
    preds=None,
    classes=None,
):
  targets, outputs, preds = (
    to_numpy(targets),
    to_numpy(outputs),
    to_numpy(preds)
  )




class Meter:
  def __init__(self):
    self.reset()

  def reset(self):
    self.max = None
    self.min = None
    self.sum = 0
    self.count = 0

  def update(self, val):
    if self.max is None or val > self.max:
      self.max = val
    if self.min is None or val < self.min:
      self.min = val

    self.sum += val
    self.count += 1

  @property
  def average(self):
    return self.sum / self.count